"""
Example 4. Finding camera parameters.
"""
import os
import argparse
import glob

import paddle
import paddle.nn as nn
import numpy as np
from skimage.io import imread, imsave
import tqdm
import imageio

import neural_renderer_paddle as nr

current_dir = os.path.dirname(os.path.realpath(__file__))
data_dir = os.path.join(current_dir, 'data')

class Model(nn.Layer):
    def __init__(self, filename_obj, filename_ref=None):
        super(Model, self).__init__()
        # load .obj
        vertices, faces = nr.load_obj(filename_obj)
        self.register_buffer('vertices', vertices[None, :, :])
        self.register_buffer('faces', faces[None, :, :])

        # create textures
        texture_size = 2
        textures = paddle.ones([1, self.faces.shape[1], texture_size, texture_size, texture_size, 3], dtype=paddle.float32)
        self.register_buffer('textures', textures)

        # load reference image
        image_ref = paddle.to_tensor((imread(filename_ref).max(-1) != 0).astype(np.float32))
        self.register_buffer('image_ref', image_ref)

        # camera parameters
        camera_position = paddle.to_tensor(np.array([6, 10, -14], dtype=np.float32))
        self.camera_position = self.create_parameter(camera_position.shape)
        self.camera_position.set_value(camera_position)

        # setup renderer
        renderer = nr.Renderer(camera_mode='look_at')
        renderer.eye = self.camera_position
        self.renderer = renderer

    def forward(self):
        image = self.renderer(self.vertices, self.faces, mode='silhouettes')
        loss = paddle.sum((image - self.image_ref[None, :, :]) ** 2)
        return loss


def make_gif(filename):
    with imageio.get_writer(filename, mode='I') as writer:
        for filename in sorted(glob.glob('/tmp/_tmp_*.png')):
            writer.append_data(imread(filename))
            os.remove(filename)
    writer.close()


def make_reference_image(filename_ref, filename_obj):
    model = Model(filename_obj)

    model.renderer.eye = nr.get_points_from_angles(2.732, 30, -15)
    images, _, _ = model.renderer.render(model.vertices, model.faces, paddle.tanh(model.textures))
    image = images.detach().cpu().numpy()[0]
    imsave(filename_ref, image)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-io', '--filename_obj', type=str, default=os.path.join(data_dir, 'teapot.obj'))
    parser.add_argument('-ir', '--filename_ref', type=str, default=os.path.join(data_dir, 'example4_ref.png'))
    parser.add_argument('-or', '--filename_output', type=str, default=os.path.join(data_dir, 'example4_result.gif'))
    parser.add_argument('-mr', '--make_reference_image', type=int, default=0)
    parser.add_argument('-g', '--gpu', type=int, default=0)
    args, _ = parser.parse_known_args()

    if args.make_reference_image:
        make_reference_image(args.filename_ref, args.filename_obj)

    model = Model(args.filename_obj, args.filename_ref)

    # optimizer = chainer.optimizers.Adam(parameters=alpha=0.1)
    optimizer = paddle.optimizer.Adam(parameters=model.parameters(), learning_rate=0.1)
    loop = tqdm.tqdm(range(1000))
    for i in loop:
        optimizer.clear_grad()
        loss = model()
        loss.backward()
        optimizer.step()
        with paddle.no_grad():
            images, _, _ = model.renderer(model.vertices, model.faces, paddle.tanh(model.textures))
        image = np.uint8(images.detach().cpu().numpy()[0].transpose(1,2,0).clip(0,1)*255)
        imsave('/tmp/_tmp_%04d.png' % i, image)
        loop.set_description('Optimizing (loss %.4f)' % loss.item())
        if loss.item() < 70:
            break
    make_gif(args.filename_output)


if __name__ == '__main__':
    main()
